require "colorize"

module Hence
  module Assembler
    extend self

    def assemble(ast : Parser::AST, opcode_names : Hash(String, UInt8), data : Data) : Bytes
      ast.body.each do |node|
        data.constants["OFFSET"] = Parser::AST::NumberArg.new(data.offset)

        case node
        when Parser::AST::ConstantDeclarationNode
          raise "Constants can't be redefined" if data.constants[node.name]?

          data.constants[node.name] = Parser::AST::BytesArg.new(node.value.resolve(data))
        when Parser::AST::LabelNode
          raise "Labels can't be redefined" if data.constants[node.name]?

          data.constants[node.name] = Parser::AST::NumberArg.new(data.offset)
        when Parser::AST::MacroCallNode
          MACROS[node.name].call(data, node.args)
        when Parser::AST::CallNode
          data.bin[data.offset] = opcode_names[node.name]
          arg = node.arg.try(&.resolve(data)) || Bytes.new(2)
          data.bin[data.offset + 1] = arg[0]
          data.bin[data.offset + 2] = arg[1]

          data.offset += 3
        else
          raise "Unexpected node type: #{typeof(node)}"
        end

        raise "Offset out of bounds: 0x#{data.offset.to_s(16).rjust(4, '0')} > 0x#{data.bin.size.to_s(16).rjust(4, '0')}" if data.offset > data.bin.size
      end

      data.bin
    end

    class RawData
      property bin
      property offset
      property constants

      def initialize(
        @bin = Bytes.new({{ 32 * 1024 }}),
        @offset = 0_u16,
        @constants = {} of String => Parser::AST::Arg
      )
      end
    end

    class Data < RawData
      property path

      def initialize(
        @path : Path,
        @bin = Bytes.new({{ 32 * 1024 }}),
        @offset = 0_u16,
        @constants = {} of String => Parser::AST::Arg
      )
      end
    end

    MACROS = {
      "debug" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        args.each do |arg|
          arg_s = arg.to_s.gsub(/\n/, "\\n")

          num = arg.resolve_as_number(data)
          slice = arg.resolve(data)

          puts arg_s.colorize(:green)
          puts "  => #{"0x#{num.to_s(16).rjust(4, '0')}".colorize(:magenta)}"
          puts "  => #{num.colorize(:magenta)}"
          puts "  => [#{slice.map { |n| "0x#{n.to_s(16).rjust(2, '0')}".colorize(:magenta) }.join(", ")}]"
          puts "  => #{"\"#{String.new(slice).gsub(/\n/, "\\n")}\"".colorize(:light_green)}"
          puts
        end
      end,
      "org" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset = args[0].resolve_as_number(data)
      end,
      "org_add" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset += args[0].resolve_as_number(data)
      end,
      "org_sub" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        data.offset -= args[0].resolve_as_number(data)
      end,
      "bytes" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        args.each do |a|
          a.resolve(data).each do |b|
            data.bin[data.offset] = b
            data.offset += 1
          end
        end
      end,
      "bw" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        string_arg = String.new(args[0].resolve(data)).gsub(/\\n/, '\n').to_slice
        pos_start = args[1]?.try(&.resolve_as_number(data)) || 0_u16
        pos_end = args[2]?.try(&.resolve_as_number(data)) || string_arg.size.to_u16

        string_arg[pos_start...pos_end].each do |b|
          data.bin[data.offset] = b
          data.offset += 1
        end
      end,
      "embed" => ->(data : Data, args : Array(Parser::AST::Arg)) do
        File.open(Path[String.new(args[0].resolve(data))].expand(data.path.dirname)) do |io|
          io.getb_to_end.each_with_index do |val, i|
            data.bin[data.offset + i] = val
          end
          data.offset += io.size
        end
      end,
      "include" => ->(_data : Data, _args : Array(Parser::AST::Arg)) do
        raise "Unimplemented"
      end,
    }
  end
end
