require "yaml"
require "ecr"

require "./firmware/*"

module Hence
  class Firmware
    include YAML::Serializable

    @[YAML::Field]
    property name

    @[YAML::Field(emit_null: true)]
    property description

    @[YAML::Field]
    property authors

    @[YAML::Field]
    property date

    @[YAML::Field]
    property version

    @[YAML::Field]
    property sections

    def initialize(
      @name : String,
      @description : String?,
      @authors : Array(String),
      @date : Time,
      @version : String,
      @sections : Array(Section)
    )
    end

    def full_name : String
      "#{@name}:#{@version}"
    end

    ECR.def_to_s "#{__DIR__}/firmware/documentation.md.ecr"

    def build_opcodes : {names: Hash(String, UInt8), opcodes: Hash(UInt8, Opcode)}
      names = {} of String => UInt8
      opcodes = {} of UInt8 => Opcode
      i = 0_u8
      @sections.each do |s|
        s.opcodes.each do |opc|
          j = if opc.opcode
                opc.opcode.not_nil!
              else
                x = i
                i += 1

                x
              end
          raise "Duplicate opcode: #{j}" if opcodes.has_key?(j)

          names[opc.name] = j
          opcodes[j] = opc.to_opcode(opc.opcode ? nil : j)
        end
      end

      {names: names, opcodes: opcodes}
    end

    def compile(microcode_names : Hash(String, UInt8), data = Assembler::RawData.new) : Bytes
      sorted_opcodes = build_opcodes[:opcodes].to_a.sort_by!(&.[0])
      data.offset = (sorted_opcodes.last[0].to_u16 + 1) * 4
      sorted_opcodes.each do |opc|
        end_offset = data.offset + opc[1].microcode.body.size * 3
        base = opc[0] * 4
        data.bin[base] = (data.offset >> 8).to_u8
        data.bin[base + 1] = (data.offset & 0xff).to_u8
        data.bin[base + 2] = (end_offset >> 8).to_u8
        data.bin[base + 3] = (end_offset & 0xff).to_u8

        i = 0_u16
        opc[1].microcode.body.each do |node|
          j = data.offset + i

          case node
          when Parser::AST::LabelNode
            data.constants[node.name] = Parser::AST::NumberArg.new(data.offset &- 3)
          when Parser::AST::CallNode
            data.bin[j] = microcode_names[node.name]
            if node.arg
              arg = node.arg.not_nil!.resolve(data)

              data.bin[j + 1] = arg.not_nil![0]
              data.bin[j + 2] = arg.not_nil![1]
            end

            i += 3
          else
            raise "Unexpected node type: #{typeof(node)}"
          end
        end

        data.offset = end_offset
      end

      data.bin
    end

    class Section
      include YAML::Serializable

      @[YAML::Field]
      property name

      @[YAML::Field(emit_null: true)]
      property description

      @[YAML::Field]
      property opcodes

      def initialize(@name : String, @description : String?, @opcodes : Array(Opcode))
      end

      class Opcode
        include YAML::Serializable

        @[YAML::Field]
        property name

        @[YAML::Field(emit_null: true)]
        property description

        @[YAML::Field(emit_null: true)]
        property example

        @[YAML::Field(emit_null: true)]
        property opcode

        @[YAML::Field(emit_null: true)]
        property arg

        @[YAML::Field]
        property stack

        @[YAML::Field]
        property microcode

        def initialize(
          @name : String,
          @description : String?,
          @example : String?,
          @opcode : UInt8?,
          @arg : String?,
          @stack : Stack,
          @microcode : String
        )
        end

        def to_opcode(opcode : UInt8? = nil) : ::Hence::Opcode
          raise "Opcode already defined" if opcode && @opcode

          Hence::Opcode.new(
            @name,
            Parser.parse(@microcode)
          )
        end

        class Stack
          include YAML::Serializable

          @[YAML::Field]
          property input

          @[YAML::Field]
          property output

          def initialize(@input : Array(String), @output : Array(String))
          end
        end
      end
    end
  end
end
