require "string_scanner"

module Hence
  module Parser
    extend self

    # ameba:disable Metrics/CyclomaticComplexity
    def parse(source : String) : AST
      s = StringScanner.new(source.gsub(/;.*|@.*|#.*/, nil))
      body = [] of AST::Node
      loop do
        s.skip(/\s*/)
        break if s.eos?

        body <<
          if constant_dirty_name = s.scan(/\S+[ \t]*=/)
            constant_name = constant_dirty_name[..-2].rstrip

            s.skip(/[ \t]*/)

            constant_values = parse_args(s).not_nil!
            raise "Constant can't have more than one value" if constant_values.size > 1

            AST::ConstantDeclarationNode.new(constant_name, constant_values.first)
          elsif label_dirty_name = s.scan(/\S+:/)
            label_name = label_dirty_name.not_nil![..-2]
            raise "Label name must be lowercase" if label_name != label_name.downcase

            AST::LabelNode.new(label_name)
            # else
            #   call_name = s.scan(/\S+/).not_nil!

            #   s.skip(/[ \t]*/)

            #   if call_name.starts_with?('.')
            #     AST::MacroCallNode.new(call_name[1..], self.parse_args(s) || [] of AST::Arg)
            #   else
            #     call_args = self.parse_args(s)
            #     raise "Op code can't be called with more than one argument" if call_args && call_args.size > 1

            #     AST::CallNode.new(call_name, call_args.try(&.first?))
            #   end
          elsif macro_name_raw = s.scan(/\.\S+/)
            s.skip(/[ \t]*/)

            AST::MacroCallNode.new(macro_name_raw[1..], parse_args(s) || [] of AST::Arg)
          else
            call_name = s.scan(/\S+/).not_nil!

            s.skip(/[ \t]*/)

            call_args = parse_args(s) || [] of AST::Arg
            raise "Op code can't be called with more than one argument" if call_args && call_args.size > 1

            AST::CallNode.new(call_name, call_args.first?)
          end
      end

      AST.new(body)
    end

    def parse_binary_operation(op : String) : AST::BinaryOperation
      case op
      when "+"
        AST::AddBinaryOperation
      when "-"
        AST::SubBinaryOperation
      when "*"
        AST::MulBinaryOperation
      when "/"
        AST::DivBinaryOperation
      when "**"
        AST::PowBinaryOperation
      else
        raise "Unexpected binary operation: #{op}"
      end.new
    end

    def parse_binary_expression_arg(scanner : StringScanner) : AST::BinaryExpressionArg?
      scanner.skip(/[ \t]*/)
      return unless scanner.scan(/\(/)

      content = scanner.scan_until(/\)/).not_nil![..-2].split(/\s/, remove_empty: true)
      raise "Malformed binary expression" if content.size < 3 || content.size % 2 == 0

      first_args = content.shift(3)
      bin_expr = AST::BinaryExpressionArg.new(
        parse_args(first_args[0] + "\n").not_nil!.first,
        parse_args(first_args[2] + "\n").not_nil!.first,
        parse_binary_operation(first_args[1])
      )

      until content.empty?
        args = content.shift(2).not_nil!
        bin_expr = AST::BinaryExpressionArg.new(
          bin_expr,
          parse_args(args[1] + "\n").not_nil!.first,
          parse_binary_operation(args[0])
        )
      end

      bin_expr
    end

    def parse_binary_expression_arg(string : String) : AST::BinaryExpressionArg?
      parse_binary_expression_arg(StringScanner.new(string))
    end

    def parse_args(scanner : StringScanner) : Array(AST::Arg)?
      args_s = scanner.scan_until(/\n/)

      if args_s.try(&.blank?)
        [] of AST::Arg
      else
        s = StringScanner.new(args_s.not_nil!.strip)
        args = [] of AST::Arg

        until s.eos?
          s.skip(/[ \t]*/)
          raise "Can't seperate arguments because of no comma" if s.skip(/,/).nil? && args.size > 0
          s.skip(/[ \t]*/)

          args <<
            if s.check(/\(/)
              parse_binary_expression_arg(s).not_nil!
            elsif string_raw = s.scan(/".*"/)
              AST::StringArg.new(string_raw[1..-2].gsub(/\\n/, '\n'))
            elsif number = s.scan(/\d[\d\w]*/)
              AST::NumberArg.new(number.to_u16(prefix: true, underscore: true, strict: true))
            else
              AST::ConstantArg.new(s.scan(/\w+/).not_nil!)
            end
        end

        args
      end
    end

    def parse_args(string : String) : Array(AST::Arg)?
      parse_args(StringScanner.new(string))
    end

    struct AST
      property body

      def initialize(@body : Body)
      end

      def to_s : String
        @body.map(&.to_s).join('\n')
      end

      abstract struct Node
        abstract def to_s : String
      end

      alias Body = Array(Node)

      struct LabelNode < Node
        property name

        def initialize(@name : String)
        end

        def to_s : String
          "#{@name}:"
        end
      end

      abstract class Arg
        abstract def resolve(data : Assembler::RawData) : Bytes

        abstract def resolve_as_number(data : Assembler::RawData) : UInt16

        abstract def to_s : String
      end

      class NumberArg < Arg
        property number

        def initialize(@number : UInt16)
        end

        def resolve(data : Assembler::RawData) : Bytes
          n = Utils.split_u16(@number)
          Bytes[n[0], n[1]]
        end

        def resolve_as_number(data : Assembler::RawData) : UInt16
          @number
        end

        def to_s : String
          @number.to_s
        end
      end

      class ConstantArg < Arg
        property name

        def initialize(@name : String)
        end

        def resolve(data : Assembler::RawData) : Bytes
          data.constants[@name].resolve(data)
        end

        def resolve_as_number(data : Assembler::RawData) : UInt16
          data.constants[@name].resolve_as_number(data)
        end

        def to_s : String
          @name
        end
      end

      class BytesArg < Arg
        property bytes

        def initialize(@bytes : Bytes)
        end

        def resolve(data : Assembler::RawData) : Bytes
          @bytes
        end

        def resolve_as_number(data : Assembler::RawData) : UInt16
          if @bytes.size == 1
            @bytes[0].to_u16
          else
            Utils.merge_u8(@bytes[0], @bytes[1])
          end
        end

        def to_s : String
          "\"#{String.new(@bytes).gsub(/\n/, "\\n")}\""
        end
      end

      class StringArg < Arg
        property string

        def initialize(@string : String)
        end

        def resolve(data : Assembler::RawData) : Bytes
          if @string.bytesize == 1
            Bytes[0_u8, @string.bytes[0]]
          else
            Bytes.new(@string.bytes.to_unsafe, @string.bytesize)
          end
        end

        def resolve_as_number(data : Assembler::RawData) : UInt16
          slice = resolve(data)

          Utils.merge_u8(slice[0], slice[1])
        end

        def to_s : String
          "\"#{@string}\""
        end
      end

      abstract struct BinaryOperation
        abstract def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg

        abstract def to_s : String
      end

      struct AddBinaryOperation < BinaryOperation
        def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg
          NumberArg.new(left.resolve_as_number(data) &+ right.resolve_as_number(data))
        end

        def to_s : String
          "+"
        end
      end

      struct SubBinaryOperation < BinaryOperation
        def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg
          NumberArg.new(left.resolve_as_number(data) &- right.resolve_as_number(data))
        end

        def to_s : String
          "-"
        end
      end

      struct MulBinaryOperation < BinaryOperation
        def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg
          NumberArg.new(left.resolve_as_number(data) &* right.resolve_as_number(data))
        end

        def to_s : String
          "*"
        end
      end

      struct DivBinaryOperation < BinaryOperation
        def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg
          NumberArg.new(left.resolve_as_number(data) // right.resolve_as_number(data))
        end

        def to_s : String
          "/"
        end
      end

      struct PowBinaryOperation < BinaryOperation
        def resolve(left : Arg, right : Arg, data : Assembler::RawData) : Arg
          NumberArg.new(left.resolve_as_number(data) &** right.resolve_as_number(data))
        end

        def to_s : String
          "**"
        end
      end

      class BinaryExpressionArg < Arg
        property left
        property right
        property op

        def initialize(@left : Arg, @right : Arg, @op : BinaryOperation)
        end

        def resolve(data : Assembler::RawData) : Bytes
          @op.resolve(@left, @right, data).resolve(data)
        end

        def resolve_as_number(data : Assembler::RawData) : UInt16
          @op.resolve(@left, @right, data).resolve_as_number(data)
        end

        def to_s : String
          # ameba:disable Lint/RedundantStringCoercion
          "(#{@left.to_s} #{@op.to_s} #{@right.to_s})"
        end
      end

      struct ConstantDeclarationNode < Node
        property name
        property value

        def initialize(@name : String, @value : Arg)
        end

        def to_s : String
          "#{@name}"
        end
      end

      struct MacroCallNode < Node
        property name
        property args

        def initialize(@name : String, @args : Array(Arg))
        end

        def to_s : String
          ".#{@name}"
        end
      end

      struct CallNode < Node
        property name
        property arg

        def initialize(@name : String, @arg : Arg? = nil)
        end

        def to_s : String
          if @arg
            # ameba:disable Lint/RedundantStringCoercion
            "#{@name} #{@arg.to_s}"
          else
            @name
          end
        end
      end
    end
  end
end
